from __future__ import annotations

import logging
import logging.handlers
import os
import sys
import types
from abc import abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from logging import CRITICAL
from logging import Logger as _Logger
from logging import _srcfile as _LOGGING_SRC_FILE

from typing_extensions import Any, Callable, Generator, Literal, cast

from .._lazy import singleton
from .._render import get_rich_exception
from ..typ._enum import LogLevel, VoidType
from ..typ.base import T
from ..typ.cls import BetterABC
from ..utils.common import find_caller_stack
from .handler import FastRotatingFileHandler, FastStreamHandler

# 取消 better-exceptions 的猴子补丁
logging._loggerClass = (  # type:ignore[attr-defined]
    logging.Logger
)


class GenericLogger(BetterABC):
    """通用日志器抽象类

    任何日志器实现本类接口，或通过 :func:`.logger_patch` 修补后，
    即可兼容 melobot 内部所有日志操作（也就可以用于 bot 初始化 :meth:`.Bot.`）
    """

    @abstractmethod
    def debug(self, msg: object) -> None:
        """`debug` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def info(self, msg: object) -> None:
        """`info` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def warning(self, msg: object) -> None:
        """`warning` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def error(self, msg: object) -> None:
        """`error` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def critical(self, msg: object) -> None:
        """`critical` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def exception(self, msg: object) -> None:
        """记录异常信息的日志"""
        raise NotImplementedError

    @abstractmethod
    def generic_lazy(
        self,
        msg: str,
        *arg_getters: Callable[[], str],
        level: LogLevel,
        with_exc: bool = False,
    ) -> None:
        """通用懒惰日志方法

        :param msg: 日志消息，可使用 %s 指定稍后填充的参数
        :param arg_getters: 填充消息 %s 位置的填充函数
        :param level: 日志等级
        :param with_exc: 是否记录异常栈信息
        """
        raise NotImplementedError

    @abstractmethod
    def generic_obj(
        self,
        msg: str,
        obj: T,
        *arg_getters: Callable[[], str],
        level: LogLevel = LogLevel.INFO,
    ) -> None:
        """通用记录对象日志方法

        :param msg: 附加的日志消息，可使用 %s 指定稍后填充的参数
        :param obj: 需要被日志记录的对象
        :param arg_getters: 填充消息 %s 位置的填充函数
        :param level: 日志等级
        """
        raise NotImplementedError


@singleton
class NullLogger(_Logger, GenericLogger):
    """空日志器，自动丢弃所有日志记录请求"""

    def __init__(self) -> None:
        super().__init__("__MELO_EMPTYLOGGER__", CRITICAL)
        self.addHandler(logging.NullHandler())

    def generic_lazy(
        self,
        msg: str,
        *arg_getters: Callable[[], str],
        level: LogLevel,
        with_exc: bool = False,
    ) -> None:
        return

    def generic_obj(
        self,
        msg: str,
        obj: T,
        *arg_getters: Callable[[], str],
        level: LogLevel = LogLevel.INFO,
    ) -> None:
        return


class Logger(_Logger, GenericLogger):
    """melobot 内置日志器

    推荐使用的日志器。实现了 :class:`GenericLogger` 接口，因此可以用于 melobot 内部日志记录

    `debug`, `info`, `warning`, `error`, `critical`, `exception`
    等接口与 :class:`logging.Logger` 用法完全一致
    """

    def findCaller(
        self, stack_info: bool = False, stacklevel: int = 1
    ) -> tuple[str, int, str, str | None]:
        *ret, sinfo = find_caller_stack(stack_info, stacklevel, is_logging_frame)
        sinfo = f"\u0000{ret[0]},{ret[2]}"
        return cast(tuple[str, int, str, str | None], (*(ret[1:]), sinfo))

    def makeRecord(self, *args: Any, **kwargs: Any) -> logging.LogRecord:
        record = super().makeRecord(*args, **kwargs)
        sinfo = record.stack_info

        *sinfo_strs, caller_info = cast(str, sinfo).split("\u0000")
        *mod_name_strs, func_lineno = caller_info.split(",")
        record.mod_name = "".join(mod_name_strs)
        record.func_lineno = int(func_lineno)

        sinfo = "".join(sinfo_strs)
        if sinfo == "":
            sinfo = None
        record.stack_info = sinfo
        return record

    def __init__(
        self,
        name: str = "[default]",
        level: LogLevel = LogLevel.INFO,
        file_level: LogLevel = LogLevel.DEBUG,
        to_console: bool = True,
        to_dir: str | None = None,
        add_tag: bool = True,
        legacy: bool = False,
        yellow_warn: bool = True,
        red_error: bool = True,
        two_stream: bool = False,
        is_parellel: bool = False,
    ) -> None:
        """初始化日志器

        :param name: 日志器的名称（唯一）
        :param level: 日志等级
        :param file_level: 日志文件的日志等级
        :param to_console: 是否输出到控制台
        :param to_dir: 保存日志文件的目录，为空则不保存文件
        :param add_tag: 记录日志时是否标识日志器名称
        :param legacy: 记录日志时是否使用传统样式（不对日志内容进行自动高亮，而是使用日志等级的五色）
        :param yellow_warn:
            记录 `LogLevel.WARN` 级别时，是否将日志内容着色为黄色。
            `legacy` 选项为 `True` 时此参数无效

        :param red_error:
            记录 `LogLevel.ERROR` 及以上级别时，是否将日志内容着色为红色。
            `legacy` 选项为 `True` 时此参数无效

        :param two_stream: 当使用记录到文件功能时，是否分离“常规日志”和“问题日志”（warning, error, critical）到不同的文件
        :param is_parellel: 日志渲染是否启用并行优化（可能导致日志小范围行间乱序）
        """
        super().__init__(name, LogLevel.DEBUG)
        self._handler_arr: list[logging.Handler] = []
        self._no_tag = not add_tag
        self._filter = _MeloLogFilter(name, yellow_warn, red_error, legacy)
        self._parellel = is_parellel

        if to_console:
            con_handler = self._add_console_handler()
            con_handler.setLevel(level)

        if to_dir is None:
            pass
        elif not two_stream:
            self._add_file_handler(to_dir, name, file_level)
        else:
            normal_handler = self._add_file_handler(to_dir, f"{name}.out", file_level)
            normal_handler.addFilter(_NormalLvlFilter(name))
            self._add_file_handler(to_dir, f"{name}.err", max(file_level, LogLevel.WARNING))

    def _add_console_handler(self) -> logging.Handler:
        fmt = self._console_fmt(self.name, self._no_tag)
        handler = self._console_handler(fmt)
        handler.addFilter(self._filter)

        self.addHandler(handler)
        self._handler_arr.append(handler)
        return handler

    def _add_file_handler(
        self, log_dir: str, name: str, level: LogLevel = LogLevel.DEBUG
    ) -> logging.Handler:
        fmt = self._file_fmt(self.name, self._no_tag)
        handler = self._file_handler(fmt, log_dir, name, level)
        handler.addFilter(self._filter)

        self.addHandler(handler)
        self._handler_arr.append(handler)
        return handler

    @staticmethod
    def _make_fmt_nocache(fmt: logging.Formatter) -> None:
        _original_format = fmt.format

        def nocache_format(record: logging.LogRecord) -> str:
            record.exc_text = None
            return _original_format(record)

        fmt.format = nocache_format  # type: ignore[method-assign]

    @staticmethod
    def _console_fmt(name: str, no_tag: bool = False) -> logging.Formatter:
        import colorlog

        fmt_arr = [
            "%(cyan)s%(asctime)s.%(msecs)03d%(reset)s",
            "%(log_color)s%(levelname)-7s%(reset)s",
            "%(blue)s%(mod_name)s%(reset)s:%(cyan)s%(func_lineno)d%(reset)s",
        ]
        if not no_tag:
            fmt_arr.insert(1, f"%(purple)s{name}%(reset)s")
        fmt_s = " | ".join(fmt_arr)
        msg_str_f = "%(log_color)s%(legacy_msg_str)s%(reset)s%(colored_msg_str)s"
        obj_str_f = "%(log_color)s%(legacy_obj)s%(reset)s%(colored_obj)s"
        fmt_s += f" - {msg_str_f}{obj_str_f}"

        fmt = colorlog.ColoredFormatter(fmt_s, datefmt="%Y-%m-%d %H:%M:%S", reset=True)
        fmt.default_msec_format = "%s.%03d"
        fmt.formatException = lambda exc_info: get_rich_exception(*exc_info)[0]  # type: ignore
        Logger._make_fmt_nocache(fmt)
        return fmt  # type: ignore[no-any-return]

    def _console_handler(self, fmt: logging.Formatter) -> logging.Handler:
        handler = FastStreamHandler(self._parellel)
        handler.setFormatter(fmt)
        return handler

    @staticmethod
    def _file_fmt(name: str, no_tag: bool = False) -> logging.Formatter:
        fmt_arr = [
            "%(asctime)s.%(msecs)03d",
            "%(levelname)-7s",
            "%(mod_name)s:%(func_lineno)d",
        ]
        if not no_tag:
            fmt_arr.insert(1, name)
        fmt_s = " | ".join(fmt_arr)
        fmt_s += " - %(msg_str)s%(obj)s"

        fmt = logging.Formatter(
            fmt=fmt_s,
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        fmt.default_msec_format = "%s.%03d"
        fmt.formatException = lambda exc_info: get_rich_exception(*exc_info)[1]  # type: ignore
        Logger._make_fmt_nocache(fmt)
        return fmt

    def _file_handler(
        self, fmt: logging.Formatter, log_dir: str, name: str, level: LogLevel
    ) -> logging.Handler:
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)

        handler = FastRotatingFileHandler(
            self._parellel,
            filename=os.path.join(log_dir, f"{name}.log"),
            maxBytes=1024 * 1024,
            backupCount=10,
            encoding="UTF-8",
        )
        handler.setLevel(level)
        handler.setFormatter(fmt)
        return handler

    def set_level(self, level: LogLevel) -> None:  # type: ignore[override]
        """设置日志等级

        日志等级自动应用于包含的所有 handler（但输出日志到文件的 handler 除外）

        :param level: 日志等级
        """
        super().setLevel(level)
        for handler in self._handler_arr:
            if not isinstance(handler, logging.handlers.RotatingFileHandler):
                handler.setLevel(level)

    def generic_lazy(
        self,
        msg: str,
        *arg_getters: Callable[[], str],
        level: LogLevel,
        with_exc: bool = False,
        stacklevel: int = 1,
    ) -> None:
        """懒惰日志方法

        :param msg: 日志消息，可使用 %s 指定稍后填充的参数
        :param arg_getters: 填充消息 %s 位置的填充函数
        :param level: 日志等级
        :param with_exc: 是否记录异常栈信息
        :param stacklevel: 打印日志时尝试解析的调用栈层级
        """
        if not self.isEnabledFor(level):
            return
        exc = sys.exc_info() if with_exc else None
        self._log(level, msg, tuple(g() for g in arg_getters), exc_info=exc, stacklevel=stacklevel)

    def generic_obj(
        self,
        msg: str,
        obj: T,
        *arg_getters: Callable[[], str],
        level: LogLevel = LogLevel.INFO,
        stacklevel: int = 1,
    ) -> None:
        """记录对象的日志方法

        :param msg: 附加的日志消息，可使用 %s 指定稍后填充的参数
        :param obj: 需要被日志记录的对象
        :param arg_getters: 填充消息 %s 位置的填充函数
        :param level: 日志等级
        :param stacklevel: 打印日志时尝试解析的调用栈层级
        """
        with self._filter.on_obj(obj):
            self.generic_lazy(msg + "\n", *arg_getters, level=level, stacklevel=stacklevel)


class _NormalLvlFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        return logging.DEBUG <= record.levelno < logging.WARNING


@dataclass
class LogInfo:
    yellow_warn: bool
    red_error: bool
    legacy: bool
    msg: str
    obj: Any


class _MeloLogFilter(logging.Filter):
    def __init__(
        self,
        name: str = "",
        yellow_warn: bool = True,
        red_error: bool = True,
        legacy: bool = False,
    ) -> None:

        super().__init__(name)
        self._obj: Any = VoidType.VOID
        self._enable_yellow_warn = yellow_warn
        self._enable_red_error = red_error
        self._legacy = legacy

    def set_obj(self, obj: Any) -> None:
        self._obj = obj

    def clear_obj(self) -> None:
        self._obj = VoidType.VOID

    @contextmanager
    def on_obj(self, obj: Any) -> Generator[None, None, None]:
        try:
            self.set_obj(obj)
            yield
        finally:
            self.clear_obj()

    def filter(self, record: logging.LogRecord) -> Literal[True]:
        msg = str(record.msg)
        if record.args:
            msg = msg % record.args
        record.msg_str = msg
        self._fill_log_info(msg, record)
        return True

    def _fill_log_info(self, msg: str, record: logging.LogRecord) -> None:
        log_info = LogInfo(
            yellow_warn=self._enable_yellow_warn,
            red_error=self._enable_red_error,
            legacy=self._legacy,
            msg=msg,
            obj=self._obj,
        )
        record.log_info = log_info


def is_logging_frame(frame: types.FrameType) -> bool:
    filename = os.path.normcase(frame.f_code.co_filename)
    return filename in (_FILE, _LOGGING_SRC_FILE) or (
        "importlib" in filename and "_bootstrap" in filename
    )


def generic_obj_meth(
    logger: GenericLogger,
    msg: str,
    obj: T,
    *arg_getters: Callable[[], str],
    level: LogLevel = LogLevel.INFO,
) -> None:
    _getters = arg_getters + (lambda: str(obj),)
    logger.generic_lazy(msg + "\n%s", *_getters, level=level)


_FILE = os.path.normcase(generic_obj_meth.__code__.co_filename)



from __future__ import annotations

import asyncio
from asyncio import create_task
from contextlib import contextmanager
from time import time_ns

from typing_extensions import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Generator,
    Generic,
    Hashable,
    Iterator,
    Literal,
    LiteralString,
    Self,
    Sequence,
    TypeVar,
    cast,
)

from ..ctx import ActionAutoExecCtx
from ..exceptions import ActionHandleError
from ..io.base import AbstractOutSource
from ..mixin import AttrReprMixin, FlagMixin
from ..typ.cls import BetterABC, abstractattr
from ..utils.common import get_id
from .content import Content

if TYPE_CHECKING:
    from .base import AbstractEchoFactory, AbstractOutputFactory


class Event(AttrReprMixin, FlagMixin):
    """事件基类

    :ivar typing.LiteralString protocol: 遵循的协议，为空则协议无关
    :ivar float time: 时间戳
    :ivar str id: id 标识
    :ivar typing.Hashable | None scope: 所在的域，可空
    :ivar typing.Sequence[Content] contents: 附加的通用内容序列
    """

    def __init__(
        self,
        protocol: LiteralString,
        time: float = -1,
        id: str = "",
        scope: Hashable | None = None,
        contents: Sequence[Content] | None = None,
    ) -> None:
        super().__init__()

        self.time = time_ns() / 1e9 if time == -1 else time
        self.id = get_id() if id == "" else id
        self.protocol = protocol
        self.contents = contents if contents else ()
        self.scope = scope

        self.spread: bool = True


EventT = TypeVar("EventT", bound=Event)


class TextEvent(Event, BetterABC):
    """文本事件类

    :ivar str text: 文本内容
    :ivar list[str] textlines: 文本分行内容
    """

    text: str = abstractattr()
    """:meta hide-value:"""

    textlines: list[str] = abstractattr()
    """:meta hide-value:"""


class Action(AttrReprMixin, FlagMixin):
    """行为基类

    :ivar float time: 时间戳
    :ivar str id: id 标识
    :ivar typing.LiteralString | None protocol: 遵循的协议，为空则协议无关
    :ivar typing.Sequence[Content] contents: 附加的通用内容序列
    :ivar typing.Hashable | None scope: 所在的域，可空
    :ivar Event | None trigger: 触发该行为的事件，为空表明不由事件触发
    """

    def __init__(
        self,
        time: float = -1,
        id: str = "",
        protocol: LiteralString | None = None,
        scope: Hashable | None = None,
        contents: Sequence[Content] | None = None,
        trigger: Event | None = None,
    ) -> None:
        super().__init__()
        self.time = time_ns() / 1e9 if time == -1 else time
        self.id = get_id() if id == "" else id
        self.protocol = protocol
        self.contents = contents if contents else ()
        self.scope = scope
        self.trigger = trigger


class Echo(AttrReprMixin, FlagMixin):
    """回应基类

    :ivar float time: 时间戳
    :ivar str id: id 标识
    :ivar typing.LiteralString | None protocol: 遵循的协议，为空则协议无关
    :ivar typing.Hashable | None scope: 所在的域，可空
    :ivar int status: 回应状态码
    :ivar str prompt: 回应提示语
    """

    def __init__(
        self,
        time: float = -1,
        id: str = "",
        protocol: LiteralString | None = None,
        scope: Hashable | None = None,
        status: int = 0,
        prompt: str = "",
    ) -> None:
        super().__init__()
        self.time = time_ns() / 1e9 if time == -1 else time
        self.id = get_id() if id == "" else id
        self.protocol = protocol
        self.scope = scope
        self.status = status
        self.prompt = prompt


ActionT = TypeVar("ActionT", bound=Action)
EchoT = TypeVar("EchoT", bound=Echo)


class ActionHandleGroup(Generic[EchoT]):
    """行为操作句柄组"""

    def __init__(self, *handles: ActionHandle[EchoT]) -> None:
        self._handles = handles

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(all: {len(self._handles)}, "
            f"done: {len(tuple(h for h in self._handles if h.status == 'DONE'))})"
        )

    def __getitem__(self, idx: int) -> ActionHandle[EchoT]:
        return self._handles[idx]

    def __len__(self) -> int:
        return len(self._handles)

    def __iter__(self) -> Iterator[ActionHandle[EchoT]]:
        return iter(self._handles)

    async def __aiter__(self) -> AsyncIterator[tuple[EchoT | None, ActionHandle[EchoT]]]:
        waits = (h._await_ret_self() for h in self._handles)
        for fut in asyncio.as_completed(waits):
            yield await fut

    async def unwrap_iter(self) -> AsyncIterator[EchoT]:
        waits = (h._await_ret_self() for h in self._handles)
        for fut in asyncio.as_completed(waits):
            echo, h = await fut
            if echo is None:
                raise ActionHandleError(
                    "迭代获取行为操作的回应失败，迭代时出现为 None 的回应", handle=h
                )
            yield echo

    def __await__(self) -> Generator[Any, Any, list[EchoT | None]]:
        return self._await_all().__await__()

    async def unwrap(self, idx: int) -> EchoT:
        handle = self._handles[idx]
        echo = await handle
        if echo is None:
            raise ActionHandleError(
                f"对行为操作 {self.unwrap.__name__} 失败，因为操作的回应为 None", handle=handle
            )
        return echo

    async def unwrap_all(self) -> list[EchoT]:
        echoes = await self._await_all()
        for idx, e in enumerate(echoes):
            if e is None:
                raise ActionHandleError(
                    f"对行为操作组 {self.unwrap_all.__name__} 失败，因为操作组的回应中有 None。",
                    handle=self._handles[idx],
                )
        return cast(list[EchoT], echoes)

    async def _await_all(self) -> list[EchoT | None]:
        echoes = await asyncio.gather(*self._handles)
        return echoes

    def execute(self) -> None:
        for h in self._handles:
            h.execute()


class ActionHandle(Generic[EchoT]):
    """行为操作句柄

    :ivar Action action: 操作包含的行为对象
    :ivar typing.Literal["PENDING", "EXECUTING", "FINISHED"] status: 操作的状态。分别对应：未执行、执行中、执行完成
    :ivar AbstractOutSource out_src: 执行操作的输出源对象
    """

    def __init__(
        self,
        action: Action,
        out_src: AbstractOutSource,
        output_factory: "AbstractOutputFactory",
        echo_factory: "AbstractEchoFactory",
    ) -> None:
        self.action = action
        self.status: Literal["PENDING", "EXECUTING", "DONE"] = "PENDING"
        self.out_src = out_src

        self._echo_fut: asyncio.Future[EchoT | None] = asyncio.Future()
        self._output_factory = output_factory
        self._echo_factory = echo_factory

        if ActionAutoExecCtx().try_get(True):
            self.execute()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(status={self.status}, fut={self._echo_fut})"

    def __await__(self) -> Generator[Any, Any, EchoT | None]:
        # 并发多次 await 是安全的，未触发执行任务时 await 也是安全的
        return self._echo_fut.__await__()

    async def _await_ret_self(self) -> tuple[EchoT | None, Self]:
        echo = await self._echo_fut
        return echo, self

    def execute(self) -> Self:
        if self.status != "PENDING":
            raise RuntimeError("行为操作正在执行或执行完毕，不应该再执行")
        self.status = "EXECUTING"
        create_task(self._execute())
        return self

    async def _execute(self) -> None:
        try:
            output_packet = await self._output_factory.create(self.action)
            echo_packet = await self.out_src.output(output_packet)
            echo = await self._echo_factory.create(echo_packet)
            self._echo_fut.set_result(echo)
        except Exception as e:
            self._echo_fut.set_exception(e)
            if isinstance(e, asyncio.CancelledError):
                raise
        finally:
            self.status = "DONE"


@contextmanager
def lazy_action() -> Generator[None, None, None]:
    """手动执行行为操作的上下文管理器

    展开一个行为操作不自动执行的上下文，适用于需要手动干预行为操作执行时机的场景
    """
    with ActionAutoExecCtx().unfold(False):
        yield



import sys
from multiprocessing import SimpleQueue
from pathlib import Path

sys.path.insert(0, Path(__file__).parent.parent.resolve().as_posix())
from tests.mp.submod import main


def test_all(argv: list[str], mod_name: str) -> None:
    if sys.argv != argv:
        raise ValueError("测试中的 argv 不匹配")
    if globals()["__name__"] != mod_name:
        raise ValueError("测试中的模块名不匹配")
    main()


def simple_test() -> None:
    main()



from typing_extensions import Any

from ..adapter.model import Event as _Event
from ..ctx import BotCtx as _BotCtx
from ..ctx import FlowCtx as _FlowCtx
from ..ctx import FlowRecord, FlowRecordStage, FlowStore
from .base import Flow, FlowNode, block, bypass, flow_to, nextn, no_deps_node, node, rewind, stop
from .register import (
    FlowDecorator,
    on_command,
    on_contain_match,
    on_end_match,
    on_event,
    on_full_match,
    on_regex_match,
    on_start_match,
    on_text,
)


def get_flow_records() -> tuple[FlowRecord, ...]:
    """获取当前上下文中的流记录

    :return: 流记录
    """
    return _FlowCtx().get_records()


def get_flow_store() -> FlowStore:
    """获取当前上下文中的流存储

    :return: 流存储
    """
    return _FlowCtx().get_store()


def get_event() -> _Event:
    """获取当前上下文中的事件

    :return: 事件
    """
    return _FlowCtx().get_event()


def try_get_event() -> _Event | None:
    """尝试获取当前上下文中的事件

    :return: 事件或空
    """
    return _FlowCtx().try_get_event()


def __getattr__(name: str) -> Any:
    if name == "f_records":
        return get_flow_records()
    elif name == "f_store":
        return get_flow_store()
    elif name == "event":
        return get_event()
    else:
        raise AttributeError


f_records: tuple[FlowRecord, ...]
"""当前上下文中的流记录"""

f_store: FlowStore
"""当前上下文中的流存储"""

event: _Event
"""当前上下文中的事件"""



from .atool import async_at, async_interval, async_later, call_at, call_later
from .base import async_guard, to_async, to_coro, to_sync
from .common import RWContext, deprecate_warn, deprecated, get_id, get_obj_name, singleton
from .deco import cooldown, if_not, lock, semaphore, speedlimit, timelimit, unfold_ctx



from __future__ import annotations

import logging
import sys
from os import PathLike
from pathlib import Path
from time import perf_counter
from types import ModuleType

from typing_extensions import Any, cast

from .._imp import ALL_EXTS, PKG_INIT_FILENAMES, ModuleLoader, SpecFinder
from ..ctx import BotCtx
from ..utils.common import find_caller_stack
from .base import GenericLogger, Logger, is_logging_frame

_SPEC_FINDER = SpecFinder()
_BOT_CTX = BotCtx()


class LogNode:
    def __init__(
        self,
        part: str,
        parent: LogNode | None = None,
    ) -> None:
        self.part = part
        self.logger: GenericLogger | None = None
        self.parent = parent
        self.children: dict[str, LogNode] = {}

    def resolve_logger(self) -> GenericLogger | None:
        if self.logger is not None:
            return self.logger
        if self.parent is not None:
            return self.parent.resolve_logger()
        return None

    def add_child(self, part: str) -> LogNode:
        if part not in self.children:
            self.children[part] = LogNode(part, self)
        return self.children[part]


class LogReflector:
    global_logger: GenericLogger | None = Logger("[global]")
    root_node = LogNode("")
    last_updated = perf_counter()

    @classmethod
    def set_global_logger(cls, logger: GenericLogger | None) -> None:
        cls.global_logger = logger

    @classmethod
    def set_module_logger(
        cls, module: str | PathLike[str] | ModuleType, logger: GenericLogger | None
    ) -> None:
        path = cls._get_mod_path(module)
        cur_node = cls.root_node
        for part in path.parts:
            cur_node = cur_node.add_child(part)
        cur_node.logger = logger
        cls.last_updated = perf_counter()

    @classmethod
    def _get_mod_path(cls, mod: str | PathLike[str] | ModuleType) -> Path:
        if isinstance(mod, str) and not mod.endswith(ALL_EXTS) and not Path(mod).exists():
            cached_mod = sys.modules.get(mod)
            if cached_mod is not None:
                return cls._get_module_path(cached_mod)

            spec = _SPEC_FINDER.find_spec(mod, paths=None)
            if spec is not None:
                loader = cast(ModuleLoader, spec.loader)
                path = loader.mb_fp
                if path.parts[-1] in PKG_INIT_FILENAMES:
                    path = path.parent
                return path

        if isinstance(mod, ModuleType):
            path = cls._get_module_path(mod)
        else:
            path = Path(mod)
            if not path.is_absolute():
                try:
                    path = path.resolve(strict=True)
                except FileNotFoundError as e:
                    raise FileNotFoundError(f"尝试解析模块路径时，发现路径 {path} 不存在") from e
        return path

    @staticmethod
    def _get_module_path(mod: ModuleType) -> Path:
        if mod.__file__ is None:
            # 此时对应没有 __init__.* 的包
            path = Path(mod.__path__[0])
        else:
            path = (
                Path(mod.__file__).parent
                if mod.__file__.endswith(PKG_INIT_FILENAMES)
                else Path(mod.__file__)
            )
        return path


def set_global_logger(logger: GenericLogger | None) -> None:
    """设置顶级域对应的日志器

    :param logger: 日志器，可以为空
    """
    LogReflector.set_global_logger(logger)


def set_module_logger(
    module: str | PathLike[str] | ModuleType, logger: GenericLogger | None
) -> None:
    """设置模块域对应的日志器

    :param module: 模块名、模块相对或绝对路径、模块对象
    :param logger: 日志器，可以为空
    """
    LogReflector.set_module_logger(module, logger)


def reflect_logger(stack_depth: int = 3) -> GenericLogger:
    _, caller_path_str, *_ = find_caller_stack(
        stacklevel=stack_depth, inner_frame_filter=is_logging_frame
    )
    try:
        caller_path = Path(caller_path_str).resolve(strict=True)
    except FileNotFoundError as e:
        raise FileNotFoundError("自动决定日志器时，尝试解析调用模块的路径失败") from e

    cur_node = LogReflector.root_node
    for part in caller_path.parts:
        cur_node = cur_node.add_child(part)
    return GenericLogProxy(cur_node)


class GenericLogProxy(GenericLogger):
    def __init__(self, node: LogNode) -> None:
        self.__log_node__ = node
        self.__cache_ts__: float = 0
        self.__cache_logger__: GenericLogger | None = None

    @property
    def __logger__(self) -> GenericLogger | None:
        return self.__get_logger__()

    def __get_logger__(self) -> GenericLogger | None:
        if LogReflector.last_updated > self.__cache_ts__:
            logger = self.__log_node__.resolve_logger()
            self.__cache_ts__ = perf_counter()
            self.__cache_logger__ = logger
        else:
            logger = self.__cache_logger__

        if logger is None:
            if bot := _BOT_CTX.try_get():
                logger = bot.logger
            if logger is None:
                logger = LogReflector.global_logger
        return logger

    def __log_meth__(self, meth_name: str, *_: Any, **kwargs: Any) -> None:
        logger = self.__logger__
        if logger is not None:
            if isinstance(logger, logging.Logger):
                kwargs["stacklevel"] = 3
            getattr(logger, meth_name)(*_, **kwargs)

    def debug(self, *_: Any, **__: Any) -> None:
        self.__log_meth__("debug", *_, **__)

    def info(self, *_: Any, **__: Any) -> None:
        self.__log_meth__("info", *_, **__)

    def warning(self, *_: Any, **__: Any) -> None:
        self.__log_meth__("warning", *_, **__)

    def error(self, *_: Any, **__: Any) -> None:
        self.__log_meth__("error", *_, **__)

    def critical(self, *_: Any, **__: Any) -> None:
        self.__log_meth__("critical", *_, **__)

    def exception(self, *_: Any, **__: Any) -> None:
        self.__log_meth__("exception", *_, **__)

    def generic_lazy(self, *_: Any, **__: Any) -> None:
        self.__log_meth__("generic_lazy", *_, **__)

    def generic_obj(self, *_: Any, **__: Any) -> None:
        self.__log_meth__("generic_obj", *_, **__)


def __getattr__(name: str) -> Any:
    if name != "logger":
        raise AttributeError
    return reflect_logger()


logger: GenericLogger



from enum import Enum
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING

from typing_extensions import Any, Callable, assert_never


class VoidType(Enum):
    """空类型，需要区别于 `None` 时使用

    .. code:: python

        # 有些时候 `None` 也是合法值，因此需要一个额外的哨兵值：
        def foo(val: Any | VoidType = VoidType.VOID) -> None:
            ...
    """

    VOID = type("_VOID", (), {})


class ExitCode(Enum):
    NORMAL = 0
    ERROR = 1
    RESTART = 2


# TODO: 考虑在最低支持 3.11 后，使用 logging.getLevelNamesMapping 兼容部分场景
class LogLevel(int, Enum):
    """日志等级枚举"""

    CRITICAL = CRITICAL
    ERROR = ERROR
    WARNING = WARNING
    INFO = INFO
    DEBUG = DEBUG


class LogicMode(Enum):
    """逻辑模式枚举类型"""

    AND = "and"
    OR = "or"
    NOT = "not"
    XOR = "xor"

    def get_operator(self) -> Callable[[Any, Any], bool]:
        match self:
            case LogicMode.AND:
                return lambda x, y: x and y
            case LogicMode.OR:
                return lambda x, y: x or y
            case LogicMode.NOT:
                return lambda x, _: not x
            case LogicMode.XOR:
                return lambda x, y: x ^ y
            case _:
                assert_never(f"不正确的逻辑类型 {self}")



import asyncio as aio
import os
from contextlib import contextmanager

import pytest as pt
import pytest_asyncio as ptaio
from pytest import fixture
from pytest_asyncio import fixture as aiofixture

from melobot._render import EXC_SHOW_INTERNAL
from melobot.log import Logger, LogLevel, set_global_logger
from melobot.log.report import set_loop_exc_log

os.environ[EXC_SHOW_INTERNAL] = "1"
set_global_logger(Logger("tests", level=LogLevel.DEBUG))
set_loop_exc_log(strict=True)
# Auto use "package" loop_scope (not pytest fixture scope) for all async test functions
pytestmark = pt.mark.asyncio(loop_scope="package")


@contextmanager
def loop_manager():
    from melobot import _run
    from melobot._lazy import singleton_clear

    manager = _run._MANAGER
    manager.started = True
    for hook in manager.started_hooks:
        hook()
    try:
        yield
    finally:
        manager.closed = True
        for hook in manager.closed_hooks:
            hook()
        singleton_clear(_run._MANAGER)
        _run._MANAGER = _run.LoopManager()



from os import listdir as _VAR1
from pathlib import Path as _VAR2
from typing import Any as _VAR3

from melobot.plugin.base import PluginInfo as _VAR8
from melobot.plugin.load import plugin_get_attr as _VAR4

_VAR5 = _VAR2(__file__).parent
_VAR6 = set(fname.split(".")[0] for fname in _VAR1(_VAR5))
_VAR7 = _VAR5.parts[-1]


def __getattr__(name: str) -> _VAR3:
    return _VAR4(_VAR7, name, _VAR6)


__plugin_info__: _VAR8
__auto_gen_version__ = "2"



import os
import shutil
import signal
import subprocess
import sys
import tempfile
import threading
from argparse import Namespace
from pathlib import Path

from typing_extensions import Any

from melobot._run import CLI_LAST_EXIT_CODE, CLI_RUN_ALIVE_FLAG, CLI_RUN_FLAG
from melobot.typ import ExitCode


def main(args: Namespace) -> None:
    if not args.entry_file.endswith(".py"):
        entry_str = f"{args.entry_file}.py"
    else:
        entry_str = args.entry_file

    entry_path = Path(entry_str)
    try:
        entry = entry_path.resolve(strict=True).as_posix()
    except FileNotFoundError:
        raise FileNotFoundError(f"不存在的入口文件：{args.entry_file}")

    cmd = [sys.executable, entry]
    cwd = Path.cwd().resolve().as_posix()
    os.environ[CLI_RUN_FLAG] = "1"
    tmp_dir = Path(tempfile.mkdtemp()).resolve()
    os.environ[CLI_RUN_ALIVE_FLAG] = str(
        tmp_dir.joinpath(f"melobot_cli_run_{threading.get_native_id()}.signal")
    )

    try:
        while True:
            retcode: int
            create_alive_sig()
            pre_handlers: list | None = None

            with subprocess.Popen(
                cmd,
                env=os.environ,
                cwd=cwd,
                stdin=sys.stdin,
                stdout=sys.stdout,
                stderr=sys.stderr,
            ) as proc:
                pre_handlers = set_signal_handler()
                retcode = proc.wait()

            if pre_handlers is not None:
                clear_signal_handler(pre_handlers)

            if CLI_LAST_EXIT_CODE in os.environ:
                os.environ.pop(CLI_LAST_EXIT_CODE)

            if retcode == ExitCode.NORMAL.value:
                break

            if retcode == ExitCode.RESTART.value:
                print("\n>>> [mb-cli] 正在重启 Bot 主程序\n")
                os.environ[CLI_LAST_EXIT_CODE] = str(ExitCode.RESTART.value)
                continue

            if retcode == ExitCode.ERROR.value:
                break

            print()
            print(f">>> [mb-cli] Bot 主程序返回了意料之外的退出码: {retcode}")
            print(">>> [mb-cli] 若提示“已安全停止运行”，则无需关注此警告")
            break

    finally:
        clear_alive_sig()
        shutil.rmtree(tmp_dir)


def create_alive_sig() -> None:
    with open(os.environ[CLI_RUN_ALIVE_FLAG], "wb"):
        pass


def clear_alive_sig() -> None:
    if os.path.exists(os.environ[CLI_RUN_ALIVE_FLAG]):
        os.remove(os.environ[CLI_RUN_ALIVE_FLAG])


def set_signal_handler() -> list:
    def signal_handler(*_: Any, **__: Any) -> None:
        # 重入安全
        clear_alive_sig()

    pre_handlers = []
    pre_handlers.append(signal.getsignal(signal.SIGINT))
    pre_handlers.append(signal.getsignal(signal.SIGTERM))
    if sys.platform == "win32":
        pre_handlers.append(signal.getsignal(signal.SIGBREAK))

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    if sys.platform == "win32":
        signal.signal(signal.SIGBREAK, signal_handler)

    return pre_handlers


def clear_signal_handler(pre_handlers: list) -> None:
    signal.signal(signal.SIGINT, pre_handlers[0])
    signal.signal(signal.SIGTERM, pre_handlers[1])
    if sys.platform == "win32":
        signal.signal(signal.SIGBREAK, pre_handlers[2])



from typing_extensions import TYPE_CHECKING

if TYPE_CHECKING:
    from .adapter.model import ActionHandle


class BotException(Exception):
    """bot 异常基类"""

    def __init__(self, *args: object):
        super().__init__(self, args)
        if not len(args):
            self.err = ""
        elif len(args) == 1:
            self.err = str(args[0])
        else:
            self.err = str(args)
        self.pretty_err = f"[{self.__class__.__module__}.{self.__class__.__qualname__}] {self.err}"

    def __str__(self) -> str:
        return self.err


class UtilError(BotException):
    """melobot.utils 异常"""


class UtilValidateError(UtilError):
    """:py:mod:`melobot.utils` 函数参数验证异常"""


class BotError(BotException):
    """melobot bot 异常"""


class SourceError(BotException):
    """melobot 源异常"""


class PluginError(BotException):
    """melobot 插件异常"""


class PluginAutoGenError(PluginError):
    """melobot 插件自动生成异常"""


class PluginLoadError(PluginError):
    """melobot 插件加载异常"""


class PluginIpcError(PluginError):
    """melobot 插件间通信异常"""


class SessionError(BotException):
    """melobot 会话异常"""


class SessionStateFailed(SessionError):
    def __init__(self, cur_state: str, meth: str) -> None:
        self.cur_state = cur_state
        super().__init__(f"当前会话状态 {cur_state} 不支持的操作：{meth}")


class SessionRuleLacked(SessionError): ...


class FlowError(BotException):
    """melobot 处理流异常"""


class LogError(BotException):
    """melobot 日志器异常"""


class HookError(BotException):
    """melobot 生命周期组件异常"""


class AdapterError(BotException):
    """melobot 适配器异常"""


class ActionHandleError(BotException):
    """melobot 行为操作异常"""

    def __init__(self, *args: object, handle: "ActionHandle") -> None:
        super().__init__(*args)
        self.handle = handle


class DependError(BotException):
    """melobot 依赖注入异常"""


class DependInitError(DependError):
    """melobot 依赖注入项初始化失败"""


class DependBindError(DependError):
    """melobot 依赖注入项值绑定失败"""


class DynamicImpError(BotException):
    """melobot 动态导入组件异常"""

    def __init__(self, *args: object, name: str | None = None, path: str | None = None) -> None:
        super().__init__(*args)
        self.name = name
        self.path = path


class DynamicImpSpecEmpty(DynamicImpError):
    """melobot 动态导入时，模块的 spec 为空"""



from typing_extensions import Any

from ..ctx import SessionCtx as _SessionCtx
from .base import Session, SessionStore, enter_session, suspend
from .option import CompareInfo, DefaultRule, Rule


def get_session() -> Session:
    """获取当前上下文中的会话

    :return: 会话
    """
    return _SessionCtx().get()


def get_session_store() -> SessionStore:
    """获取当前上下文中的会话存储

    :return: 会话存储
    """
    return _SessionCtx().get().store


def get_rule() -> Rule:
    """获取当前上下文中的会话规则

    :return: 会话规则
    """
    return _SessionCtx().get_rule()


def __getattr__(name: str) -> Any:
    if name == "session":
        return get_session()
    elif name == "s_store":
        return get_session_store()
    elif name == "rule":
        return get_rule()
    else:
        raise AttributeError


session: Session
"""当前上下文中的会话"""

s_store: SessionStore
"""当前上下文中的会话存储"""

rule: Rule
"""当前上下文中的会话规则"""



from typing_extensions import Callable, Sequence

from melobot.handle import FlowDecorator
from melobot.session import DefaultRule, Rule
from melobot.typ import SyncOrAsyncCallable
from melobot.utils.check import Checker, checker_join
from melobot.utils.match import Matcher
from melobot.utils.parse import Parser

from .adapter.event import Event, MessageEvent, MetaEvent, NoticeEvent, RequestEvent
from .utils import check


def on_event(
    checker: Checker | None | SyncOrAsyncCallable[[Event], bool] = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    rule: Rule[Event] | None = None,
) -> FlowDecorator:
    return FlowDecorator(
        checker=checker_join(lambda e: isinstance(e, Event), checker),  # type: ignore[arg-type]
        priority=priority,
        block=block,
        temp=temp,
        decos=decos,
        rule=rule,  # type: ignore[arg-type]
    )


def on_message(
    checker: Checker | None | SyncOrAsyncCallable[[MessageEvent], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    return FlowDecorator(
        checker=checker_join(lambda e: isinstance(e, MessageEvent), checker),  # type: ignore[arg-type]
        matcher=matcher,
        parser=parser,
        priority=priority,
        block=block,
        temp=temp,
        decos=decos,
        rule=DefaultRule() if legacy_session else None,
    )


def on_at_qq(
    qid: int | None = None,
    checker: Checker | None | SyncOrAsyncCallable[[MessageEvent], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    _checker = check.AtMsgChecker(qid if qid is not None else "all")
    return FlowDecorator(
        checker=checker_join(lambda e: isinstance(e, Event), _checker, checker),  # type: ignore[arg-type]
        matcher=matcher,
        parser=parser,
        priority=priority,
        block=block,
        temp=temp,
        decos=decos,
        rule=DefaultRule() if legacy_session else None,
    )


def on_request(
    checker: Checker | None | SyncOrAsyncCallable[[RequestEvent], bool] = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    rule: Rule[Event] | None = None,
) -> FlowDecorator:
    return FlowDecorator(
        checker=checker_join(lambda e: isinstance(e, RequestEvent), checker),  # type: ignore[arg-type]
        priority=priority,
        block=block,
        temp=temp,
        decos=decos,
        rule=rule,  # type: ignore[arg-type]
    )


def on_notice(
    checker: Checker | None | SyncOrAsyncCallable[[NoticeEvent], bool] = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    rule: Rule[Event] | None = None,
) -> FlowDecorator:
    return FlowDecorator(
        checker=checker_join(lambda e: isinstance(e, NoticeEvent), checker),  # type: ignore[arg-type]
        priority=priority,
        block=block,
        temp=temp,
        decos=decos,
        rule=rule,  # type: ignore[arg-type]
    )


def on_meta(
    checker: Checker | None | SyncOrAsyncCallable[[MetaEvent], bool] = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    rule: Rule[Event] | None = None,
) -> FlowDecorator:
    return FlowDecorator(
        checker=checker_join(lambda e: isinstance(e, MetaEvent), checker),  # type: ignore[arg-type]
        priority=priority,
        block=block,
        temp=temp,
        decos=decos,
        rule=rule,  # type: ignore[arg-type]
    )



import asyncio
import hmac
import json
import time
from asyncio import Future

import aiohttp
import aiohttp.web

from melobot.io import SourceLifeSpan
from melobot.log import LogLevel, log_exc, logger

from .base import BaseIOSource
from .packet import EchoPacket, InPacket, OutPacket


class HttpIO(BaseIOSource):
    def __init__(
        self,
        onebot_url: str,
        serve_host: str,
        serve_port: int,
        secret: str | None = None,
        access_token: str | None = None,
        cd_time: float = 0,
    ) -> None:
        super().__init__(cd_time)
        self.onebot_url = onebot_url
        self.host: str = serve_host
        self.port: int = serve_port
        self.serve_site: aiohttp.web.TCPSite
        self.client_session: aiohttp.ClientSession
        self.secret = secret
        self.access_token = access_token

        self._tasks: list[asyncio.Task] = []
        self._in_buf: asyncio.Queue[InPacket] = asyncio.Queue()
        self._out_buf: asyncio.Queue[OutPacket] = asyncio.Queue()
        self._pre_send_time = time.time_ns()

        self._echo_table: dict[str, tuple[str, Future[EchoPacket]]] = {}
        self._opened = asyncio.Event()
        self._lock = asyncio.Lock()

    async def _respond(self, request: aiohttp.web.Request) -> aiohttp.web.Response:
        if not self._opened.is_set():
            self._opened.set()

        data = await request.content.read()
        if data == b"":
            return aiohttp.web.Response(status=400)

        if self.secret is not None:
            sign = hmac.new(self.secret.encode(), data, "sha1").hexdigest()
            recv_sign = request.headers["X-Signature"][len("sha1=") :]

            if sign != recv_sign:
                logger.warning("OneBot 实现程序鉴权不通过，本次上报数据将不会被处理")
                logger.generic_obj("试图上报的数据", data, level=LogLevel.WARNING)
                return aiohttp.web.Response(status=403)

        try:
            raw = json.loads(data.decode())
            if (
                self._hook_bus.get_evoke_time(SourceLifeSpan.STARTED) != -1
                and raw.get("post_type") == "meta_event"
                and raw.get("meta_event_type") == "lifecycle"
                and raw.get("sub_type") == "connect"
            ):
                await self._hook_bus.emit(SourceLifeSpan.RESTARTED, False)
            logger.generic_obj("收到上报，未格式化的字典", str(raw), level=LogLevel.DEBUG)
            await self._in_buf.put(InPacket(time=raw["time"], data=raw))

        except Exception as e:
            log_exc(e, msg="OneBot v11 HTTP IO 源输入异常", obj=raw)

        return aiohttp.web.Response(status=204)

    async def _output_loop(self) -> None:
        while True:
            try:
                await self._opened.wait()

                out_packet = await self._out_buf.get()
                wait_time = self.cd_time - ((time.time_ns() - self._pre_send_time) / 1e9)
                await asyncio.sleep(wait_time)
                asyncio.create_task(self._handle_output(out_packet))
                self._pre_send_time = time.time_ns()

            except asyncio.CancelledError:
                raise

            except Exception as e:
                log_exc(e, msg="OneBot v11 HTTP IO 源输出异常", obj=locals())

    async def _handle_output(self, packet: OutPacket) -> None:
        try:
            headers: dict | None = None
            if self.access_token is not None:
                headers = {"Authorization": f"Bearer {self.access_token}"}

            http_resp = await self.client_session.post(
                f"{self.onebot_url}/{packet.action_type}",
                json=packet.action_params,
                headers=headers,
            )
            if packet.echo_id is None:
                return

            raw = await http_resp.json()
            echo_id = raw.get("echo")
            if echo_id in (None, ""):
                return

            action_type, fut = self._echo_table.pop(echo_id)
            fut.set_result(
                EchoPacket(
                    time=int(time.time()),
                    data=raw,
                    ok=raw["status"] == "ok",
                    status=raw["retcode"],
                    action_type=action_type,
                )
            )
        except aiohttp.ContentTypeError:
            logger.error("OneBot v11 HTTP IO 源无法解析上报数据。可能是 access_token 未配置或错误")
        except Exception as e:
            log_exc(e, msg="OneBot v11 HTTP IO 源输出异常", obj=packet.data)

    async def open(self) -> None:
        if self.opened():
            return

        async with self._lock:
            if self.opened():
                return

            self.client_session = aiohttp.ClientSession()
            app = aiohttp.web.Application()
            app.add_routes([aiohttp.web.post("/", self._respond)])
            runner = aiohttp.web.AppRunner(app)

            await runner.setup()
            self.serve_site = aiohttp.web.TCPSite(runner, self.host, self.port)
            await self.serve_site.start()
            self._tasks.append(asyncio.create_task(self._output_loop()))

            logger.info("OneBot v11 HTTP IO 源就绪，等待实现端上线中")
            await self._opened.wait()
            logger.info("OneBot v11 HTTP IO 源双向通信已建立")

    def opened(self) -> bool:
        return self._opened.is_set()

    async def close(self) -> None:
        if not self.opened():
            return

        async with self._lock:
            if not self.opened():
                return

            await self.serve_site.stop()
            await self.client_session.close()
            for t in self._tasks:
                t.cancel()
            if len(self._tasks):
                await asyncio.wait(self._tasks)
            self._tasks.clear()

            self._opened.clear()
            logger.info("OneBot v11 HTTP IO 源已停止运行")

    async def input(self) -> InPacket:
        return await self._in_buf.get()

    async def output(self, packet: OutPacket) -> EchoPacket:
        await self._out_buf.put(packet)
        if packet.echo_id is None:
            return EchoPacket(noecho=True)

        fut: Future[EchoPacket] = asyncio.get_running_loop().create_future()
        self._echo_table[packet.echo_id] = (packet.action_type, fut)
        return await fut



from __future__ import annotations

import asyncio
import os
import signal
import sys
from asyncio import get_running_loop
from weakref import WeakSet

from typing_extensions import Any, Callable, Coroutine, NoReturn

from ._lazy import singleton, singleton_clear
from .typ._enum import ExitCode

CLI_RUN_FLAG = "MELOBOT_CLI_RUN"
CLI_RUN_ALIVE_FLAG = "MELOBOT_CLI_RUN_ALIVE"
CLI_LAST_EXIT_CODE = "MELOBOT_CLI_LAST_EXIT_CODE"


@singleton
class LoopManager:
    def __init__(self) -> None:
        self.root_task: asyncio.Task | None = None

        self.started = False
        self.closed = False
        self.stop_accepted = False
        self._next_manager: LoopManager

        self.started_hooks: set[Callable[[], Any]] = set()
        self.closed_hooks: set[Callable[[], Any]] = set()
        self.immunity_tasks: WeakSet[asyncio.Task] = WeakSet()

    def run(self, root: Coroutine[Any, Any, None], debug: bool) -> None:
        try:
            # TODO: 在升级最低支持到 3.11 后，考虑更换为 new_event_loop，并使用 asyncio.Runner 来运行
            loop = asyncio.get_event_loop()
            asyncio.get_event_loop_policy().set_event_loop(loop)

            # TODO: 在升级最低支持到 3.12 后，设置为默认标准
            if sys.version_info >= (3, 12):
                loop.set_task_factory(asyncio.eager_task_factory)

            loop.set_debug(debug)
            signal.signal(signal.SIGINT, self.stop)
            signal.signal(signal.SIGTERM, self.stop)
            if sys.platform == "win32":
                signal.signal(signal.SIGBREAK, self.stop)

            main = self._loop_main(root)
            loop.run_until_complete(main)

        except asyncio.CancelledError:
            pass

        finally:
            try:
                self._loop_cancel_all(loop)
                loop.run_until_complete(loop.shutdown_asyncgens())
                loop.run_until_complete(loop.shutdown_default_executor())
            finally:
                loop.set_exception_handler(None)
                asyncio.get_event_loop_policy().set_event_loop(None)
                loop.close()
                self._prepare_next_works()

    def _prepare_next_works(self) -> None:
        global _MANAGER
        singleton_clear(self)
        _MANAGER = LoopManager()
        _MANAGER.__dict__.update(self._next_manager.__dict__)
        _MANAGER._next_manager = _MANAGER.__class__()

    async def _loop_main(self, root: Coroutine[Any, Any, None]) -> None:
        self.started = True
        self.root_task = asyncio.create_task(root)
        for hook in self.started_hooks:
            hook()

        if CLI_RUN_FLAG in os.environ:
            while True:
                if self.stop_accepted or self.root_task.done():
                    break
                if not os.path.exists(os.environ[CLI_RUN_ALIVE_FLAG]):
                    self.root_task.cancel()
                    break
                await asyncio.sleep(0.45)
        else:
            await self.root_task

    def _loop_cancel_all(self, loop: asyncio.AbstractEventLoop) -> None:
        self.closed = True
        for hook in self.closed_hooks:
            hook()

        to_cancel = asyncio.all_tasks(loop)
        if not to_cancel:
            return
        for task in to_cancel:
            if task not in self.immunity_tasks:
                task.cancel()
        loop.run_until_complete(asyncio.tasks.gather(*to_cancel, return_exceptions=True))

        for task in to_cancel:
            if task.cancelled():
                continue
            if task.exception() is not None:
                loop.call_exception_handler(
                    {
                        "message": "事件循环关闭时，抛出未捕获的异常",
                        "exception": task.exception(),
                        "task": task,
                    }
                )

    def stop(self, *_: Any, **__: Any) -> None:
        if self.stop_accepted:
            return
        self.stop_accepted = True
        if self.root_task is not None and not self.root_task.done():
            get_running_loop().call_soon(self.root_task.cancel)

    def restart(self) -> NoReturn:
        sys.exit(ExitCode.RESTART.value)

    def is_from_restart(self) -> bool:
        return (
            CLI_LAST_EXIT_CODE in os.environ
            and int(os.environ[CLI_LAST_EXIT_CODE]) == ExitCode.RESTART.value
        )

    def is_restartable(self) -> bool:
        if CLI_RUN_FLAG in os.environ:
            return True
        return False


_MANAGER = LoopManager()
_MANAGER._next_manager = _MANAGER.__class__()


def register_loop_started_hook(func: Callable[[], Any], allow_next: bool = False) -> None:
    if _MANAGER.started and not _MANAGER.closed:
        raise RuntimeError("事件循环正在运行，无法添加新的事件循环启动 hook")
    if _MANAGER.closed:
        if allow_next:
            _MANAGER._next_manager.started_hooks.add(func)
        else:
            raise RuntimeError("事件循环已关闭，无法添加新的事件循环启动 hook")
    _MANAGER.started_hooks.add(func)


def register_loop_closed_hook(func: Callable[[], Any], allow_next: bool = False) -> None:
    if _MANAGER.closed:
        if allow_next:
            _MANAGER._next_manager.closed_hooks.add(func)
        else:
            raise RuntimeError("事件循环已关闭，无法添加新的事件循环关闭 hook")
    _MANAGER.closed_hooks.add(func)


def add_immunity_task(task: asyncio.Task) -> asyncio.Task:
    if _MANAGER.closed:
        raise RuntimeError("事件循环已关闭，无法添加新的取消豁免任务")
    _MANAGER.immunity_tasks.add(task)
    return task


def is_async_running() -> bool:
    return _MANAGER.started and not _MANAGER.closed


def set_loop_exc_handler(
    exc_handler: Callable[[asyncio.AbstractEventLoop, dict[str, Any]], Any],
) -> None:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        register_loop_started_hook(
            lambda: asyncio.get_running_loop().set_exception_handler(exc_handler)
        )
    else:
        loop.set_exception_handler(exc_handler)



import asyncio
from functools import partial
from logging import ERROR, WARNING, LogRecord, StreamHandler
from logging.handlers import RotatingFileHandler
from os import PathLike

from typing_extensions import TYPE_CHECKING, Any, Callable, cast

from .._lazy import singleton, singleton_clear
from .._render import get_rich_object, get_rich_repr
from .._run import (
    add_immunity_task,
    is_async_running,
    register_loop_closed_hook,
    register_loop_started_hook,
)
from ..typ import P, T, VoidType

if TYPE_CHECKING:
    from .base import LogInfo


class FastStreamHandler(StreamHandler):
    def __init__(self, is_parellel: bool) -> None:
        super().__init__()
        self.render = RecordRender(is_parellel)

    def emit(self, record: LogRecord) -> None:
        if is_async_running():
            t = asyncio.create_task(self.render.async_format(record))
            add_immunity_task(t)
            t.add_done_callback(partial(_format_cb, record, super(FastStreamHandler, self).emit))
        else:
            self.render.sync_format(record)
            super().emit(record)


class FastRotatingFileHandler(RotatingFileHandler):
    def __init__(
        self,
        is_parellel: bool,
        filename: str | PathLike[str],
        mode: str = "a",
        maxBytes: int = 0,
        backupCount: int = 0,
        encoding: str | None = None,
        delay: bool = False,
        errors: str | None = None,
    ) -> None:
        super().__init__(filename, mode, maxBytes, backupCount, encoding, delay, errors)
        self.render = RecordRender(is_parellel)

    def emit(self, record: LogRecord) -> None:
        if is_async_running():
            t = asyncio.create_task(self.render.async_format(record))
            add_immunity_task(t)
            t.add_done_callback(
                partial(_format_cb, record, super(FastRotatingFileHandler, self).emit)
            )
        else:
            self.render.sync_format(record)
            super().emit(record)


@singleton
def get_exc_report_func() -> Callable[[BaseException, str], None]:
    from .report import log_exc

    return log_exc


def _format_cb(
    record: LogRecord, super_emit: Callable[[LogRecord], None], t: asyncio.Task[None]
) -> None:
    exc = t.exception()
    if exc is not None:
        get_exc_report_func()(exc, f"日志格式化渲染出现异常，内容：{record.msg}")
        return
    super_emit(record)


@singleton
class LogRenderRunner:
    def __init__(self) -> None:
        from .. import _render
        from ..mp import ProcessPoolExecutor  # noqa: F811

        self.pool = ProcessPoolExecutor(_render.__file__, max_workers=1)
        self.task_q = asyncio.Queue[tuple[Callable, asyncio.Future, Any, Any]]()
        self.ref_pairs: list[tuple[Any, str]] = []
        self.done = asyncio.Event()

        register_loop_closed_hook(self._mark_done, allow_next=True)
        if is_async_running():
            add_immunity_task(asyncio.create_task(self._render_loop()))
        else:
            register_loop_started_hook(
                lambda: add_immunity_task(asyncio.create_task(self._render_loop())), allow_next=True
            )

    def add_ref(self, obj: Any, attr_name: str) -> None:
        self.ref_pairs.append((obj, attr_name))

    async def run(self, func: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        fut = asyncio.get_running_loop().create_future()
        await self.task_q.put((func, fut, args, kwargs))
        res = await fut
        return cast(T, res)

    def _mark_done(self) -> None:
        self.done.set()

    async def _render_loop(self) -> None:
        # 要尽可能将队列里所有日志都渲染完，而不是直接取消
        loop = asyncio.get_running_loop()
        done_task = add_immunity_task(asyncio.create_task(self.done.wait()))

        with self.pool:
            while True:
                # 想象一种情况：q_task 完成，随后日志渲染（发生了 await），回来时 done_task 已完成
                if done_task.done():
                    break
                q_task = add_immunity_task(asyncio.create_task(self.task_q.get()))
                await asyncio.wait((q_task, done_task), return_when=asyncio.FIRST_COMPLETED)
                if done_task.done():
                    if not q_task.done():
                        q_task.cancel()
                    else:
                        self.task_q.put_nowait(q_task.result())
                    break
                else:
                    func, fut, args, kwargs = q_task.result()
                    wrapped_func = partial(func, **kwargs)

                try:
                    res = await loop.run_in_executor(self.pool, wrapped_func, *args)
                except Exception as e:
                    fut.set_exception(e)
                else:
                    fut.set_result(res)

            # 收尾工作，此时把队列的任务消耗完即可
            # 也有可能此时再增加任务，但是至于能不能处理已经不重要
            # 因为至少 done_task 完成前的任务都处理了，这就足够了
            while self.task_q.qsize():
                func, fut, args, kwargs = self.task_q.get_nowait()
                wrapped_func = partial(func, **kwargs)
                try:
                    res = await loop.run_in_executor(self.pool, wrapped_func, *args)
                except Exception as e:
                    fut.set_exception(e)
                else:
                    fut.set_result(res)

        singleton_clear(self)
        new_runner = LogRenderRunner()
        for obj, attr_name in self.ref_pairs:
            try:
                setattr(obj, attr_name, new_runner)
                new_runner.add_ref(obj, attr_name)
            except Exception as e:
                get_exc_report_func()(e, f"清理 {self.__class__.__name__} 引用时发生异常")


class RecordRender:
    def __init__(self, is_parellel: bool = True) -> None:
        from rich.style import Style

        self.runner = LogRenderRunner() if is_parellel else None
        if self.runner is not None:
            self.runner.add_ref(self, "runner")

        self._yellow_style = Style(color="yellow")
        self._red_style = Style(color="red")

    def sync_format(self, record: LogRecord) -> None:
        log_info = cast("LogInfo", record.log_info)  # type: ignore[attr-defined]

        yellow_style = self._yellow_style
        red_style = self._red_style
        yellow_warn = log_info.yellow_warn
        red_error = log_info.red_error
        legacy = log_info.legacy
        msg = log_info.msg
        obj = log_info.obj

        if legacy:
            record.legacy_msg_str, record.colored_msg_str, record.msg_str = msg, "", msg

            if obj is VoidType.VOID:
                record.legacy_obj, record.obj = "", ""
            else:
                record.legacy_obj = record.obj = get_rich_object(obj, no_color=True)[1]
            record.colored_obj = ""
            return

        record.legacy_msg_str = ""
        record.legacy_obj = ""

        if red_error and record.levelno >= ERROR:
            record.colored_msg_str, record.msg_str = get_rich_repr(msg, red_style)
        elif yellow_warn and ERROR > record.levelno >= WARNING:
            record.colored_msg_str, record.msg_str = get_rich_repr(msg, yellow_style)
        else:
            record.colored_msg_str, record.msg_str = get_rich_repr(msg)

        if obj is VoidType.VOID:
            record.colored_obj, record.obj = "", ""
        elif red_error and record.levelno >= ERROR:
            record.legacy_obj = record.obj = get_rich_object(obj, no_color=True)[1]
            record.colored_obj = ""
        elif yellow_warn and ERROR > record.levelno >= WARNING:
            record.legacy_obj = record.obj = get_rich_object(obj, no_color=True)[1]
            record.colored_obj = ""
        else:
            record.colored_obj, record.obj = get_rich_object(obj)

    async def async_format(self, record: LogRecord) -> None:
        if self.runner is None:
            self.sync_format(record)
            return

        log_info = cast("LogInfo", record.log_info)  # type: ignore[attr-defined]

        yellow_style = self._yellow_style
        red_style = self._red_style
        yellow_warn = log_info.yellow_warn
        red_error = log_info.red_error
        legacy = log_info.legacy
        msg = log_info.msg
        obj = (
            self._to_easy_serialize(log_info.obj)
            if log_info.obj is not VoidType.VOID
            else VoidType.VOID
        )

        if legacy:
            record.legacy_msg_str, record.colored_msg_str, record.msg_str = msg, "", msg

            if obj is VoidType.VOID:
                record.legacy_obj, record.obj = "", ""
            else:
                record.legacy_obj = record.obj = (
                    await self.runner.run(get_rich_object, obj, no_color=True)
                )[1]
            record.colored_obj = ""
            return

        record.legacy_msg_str = ""
        record.legacy_obj = ""

        if red_error and record.levelno >= ERROR:
            record.colored_msg_str, record.msg_str = await self.runner.run(
                get_rich_repr, msg, red_style
            )
        elif yellow_warn and ERROR > record.levelno >= WARNING:
            record.colored_msg_str, record.msg_str = await self.runner.run(
                get_rich_repr, msg, yellow_style
            )
        else:
            record.colored_msg_str, record.msg_str = await self.runner.run(get_rich_repr, msg)

        if obj is VoidType.VOID:
            record.colored_obj, record.obj = "", ""
        elif red_error and record.levelno >= ERROR:
            record.legacy_obj = record.obj = (
                await self.runner.run(get_rich_object, obj, no_color=True)
            )[1]
            record.colored_obj = ""
        elif yellow_warn and ERROR > record.levelno >= WARNING:
            record.legacy_obj = record.obj = (
                await self.runner.run(get_rich_object, obj, no_color=True)
            )[1]
            record.colored_obj = ""
        else:
            record.colored_obj, record.obj = await self.runner.run(get_rich_object, obj)

    def _to_easy_serialize(self, obj: Any) -> Any:
        if isinstance(obj, str):
            return obj
        if isinstance(obj, (bytes, int, float, complex, bytearray)):
            return obj
        if obj in (None, True, False, Ellipsis, NotImplemented):
            return obj
        if isinstance(obj, (tuple, list, set)):
            return type(obj)(map(self._to_easy_serialize, obj))
        if isinstance(obj, dict):
            return {self._to_easy_serialize(k): self._to_easy_serialize(v) for k, v in obj.items()}
        return str(obj)



from __future__ import annotations

from enum import Enum

from typing_extensions import Iterable, Literal, Optional, cast

from melobot.typ import SyncOrAsyncCallable
from melobot.utils.check import Checker

from ..adapter.event import Event, GroupMessageEvent, MessageEvent
from ..adapter.segment import AtSegment


class LevelRole(int, Enum):
    """用户权限等级枚举"""

    OWNER = 1 << 4
    SU = 1 << 3
    WHITE = 1 << 2
    NORMAL = 1 << 1
    BLACK = 1


class GroupRole(int, Enum):
    """群权限等级枚举"""

    OWNER = 1 << 2
    ADMIN = 1 << 1
    MEMBER = 1
    NOT_IN_GROUP = 0


def get_level_role(checker: MsgChecker, event: MessageEvent) -> LevelRole:
    """获得消息事件对应的分级权限等级

    :param event: 消息事件
    :return: 分级权限等级
    """
    _this = get_level_role
    _flag = (checker.black_users, checker.owner, checker.super_users, checker.white_users)
    ret = event.flag_get(_this, _flag, raise_exc=False, default=None)
    if ret is not None:
        return cast(LevelRole, ret)

    qid = event.user_id
    if qid in _flag[0]:
        res = LevelRole.BLACK
    elif qid == _flag[1]:
        res = LevelRole.OWNER
    elif qid in _flag[2]:
        res = LevelRole.SU
    elif qid in _flag[3]:
        res = LevelRole.WHITE
    else:
        res = LevelRole.NORMAL
    event.flag_set(_this, _flag, res)
    return res


def get_group_role(event: MessageEvent) -> GroupRole:
    """获得消息事件对应的群权限等级

    :param event: 消息事件
    :return: 群权限等级
    """
    _this = get_group_role
    ret = event.flag_get(_this, _this, raise_exc=False, default=None)
    if ret is not None:
        return cast(GroupRole, ret)

    if not event.is_group():
        res = cast(GroupRole, GroupRole.NOT_IN_GROUP)
    elif event.sender.is_group_owner():
        res = GroupRole.OWNER
    elif event.sender.is_group_admin():
        res = GroupRole.ADMIN
    else:
        res = GroupRole.MEMBER
    event.flag_set(_this, _this, res)
    return res


class MsgChecker(Checker[Event]):
    """消息事件分级权限检查器

    主要分 主人、超级用户、白名单用户、普通用户、黑名单用户 五级
    """

    def __init__(
        self,
        role: LevelRole | GroupRole,
        owner: Optional[int] = None,
        super_users: Optional[Iterable[int]] = None,
        white_users: Optional[Iterable[int]] = None,
        black_users: Optional[Iterable[int]] = None,
        fail_cb: Optional[SyncOrAsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个消息事件分级权限检查器

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号
        :param white_users: 白名单用户 qq 号
        :param black_users: 黑名单用户 qq 号
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(fail_cb)
        self.check_role = role

        self.owner = owner
        self.super_users = tuple(super_users) if super_users is not None else ()
        self.white_users = tuple(white_users) if white_users is not None else ()
        self.black_users = tuple(black_users) if black_users is not None else ()
        self._hash_tag: tuple[int | None | tuple, ...] = (
            self.check_role,
            self.owner,
            self.super_users,
            self.white_users,
            self.black_users,
        )

    def _check(self, event: MessageEvent) -> bool:
        e_level: LevelRole | GroupRole
        if isinstance(self.check_role, LevelRole):
            e_level = get_level_role(self, event)
            status = LevelRole.BLACK < e_level and e_level >= self.check_role
        else:
            e_level = get_group_role(event)
            status = e_level >= self.check_role
        return status

    async def check(self, event: Event) -> bool:
        status = is_msg = False
        try:
            ret, is_msg = event.flag_get(
                self.__class__, self._hash_tag, raise_exc=False, default=(None, False)
            )
            if ret is not None:
                return cast(bool, ret)

            # 不要使用 isinstace，避免通过反射模式注入的 event 依赖产生误判结果
            if not event.is_message():
                status = is_msg = False
            else:
                is_msg = True
                status = self._check(cast(MessageEvent, event))

            event.flag_set(self.__class__, self._hash_tag, (status, is_msg))
            return status

        finally:
            if not status and is_msg and self.fail_cb is not None:
                await self.fail_cb()


class GroupMsgChecker(MsgChecker):
    """群聊消息事件分级权限检查器

    基本功能与 :class:`MsgChecker` 一致。但增加了白名单机制，
    不提供 `white_groups` 参数默认拒绝所有群聊消息事件。

    对所有私聊消息事件校验不通过。
    """

    def __init__(
        self,
        role: LevelRole | GroupRole,
        owner: Optional[int] = None,
        super_users: Optional[Iterable[int]] = None,
        white_users: Optional[Iterable[int]] = None,
        black_users: Optional[Iterable[int]] = None,
        white_groups: Optional[Iterable[int]] = None,
        fail_cb: Optional[SyncOrAsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个群聊消息事件分级权限检查器

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号
        :param white_users: 白名单用户 qq 号
        :param black_users: 黑名单用户 qq 号
        :param white_groups: 白名单群号（不在其中的群不通过校验）
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(role, owner, super_users, white_users, black_users, fail_cb)
        self.white_group_list = tuple(white_groups) if white_groups is not None else ()
        self._hash_tag = (*self._hash_tag, self.white_group_list)

    def _check(self, event: MessageEvent) -> bool:
        # 不要使用 isinstace，避免通过反射模式注入的 event 依赖产生误判结果
        if event.is_private():
            return False
        if len(self.white_group_list) == 0:
            return False
        if cast(GroupMessageEvent, event).group_id not in self.white_group_list:
            return False
        return super()._check(event)


class PrivateMsgChecker(MsgChecker):
    """私聊消息事件分级权限检查器

    基本功能与 :class:`MsgChecker` 一致。

    对所有群聊消息事件校验不通过。
    """

    def __init__(
        self,
        role: LevelRole,
        owner: Optional[int] = None,
        super_users: Optional[Iterable[int]] = None,
        white_users: Optional[Iterable[int]] = None,
        black_users: Optional[Iterable[int]] = None,
        fail_cb: Optional[SyncOrAsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个私聊消息事件分级权限检查器

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号
        :param white_users: 白名单用户 qq 号
        :param black_users: 黑名单用户 qq 号
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(role, owner, super_users, white_users, black_users, fail_cb)

    def _check(self, event: MessageEvent) -> bool:
        if not event.is_private():
            return False
        return super()._check(event)


class MsgCheckerFactory:
    """消息事件分级权限检查器的工厂

    预先存储检查依据（各等级数据），指定检查等级后，可返回一个 :class:`MsgChecker` 类的对象
    """

    def __init__(
        self,
        owner: Optional[int] = None,
        super_users: Optional[Iterable[int]] = None,
        white_users: Optional[Iterable[int]] = None,
        black_users: Optional[Iterable[int]] = None,
        white_groups: Optional[Iterable[int]] = None,
        fail_cb: Optional[SyncOrAsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个消息事件分级权限检查器的工厂

        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号
        :param white_users: 白名单用户 qq 号
        :param black_users: 黑名单用户 qq 号
        :param white_groups: 白名单群号（不在其中的群不通过校验）
        :param fail_cb: 检查不通过的回调（这将自动附加到生成的检查器上）
        """
        self.owner = owner
        self.super_users = tuple(super_users) if super_users is not None else ()
        self.white_users = tuple(white_users) if white_users is not None else ()
        self.black_users = tuple(black_users) if black_users is not None else ()
        self.white_groups = tuple(white_groups) if white_groups is not None else ()

        self.fail_cb = fail_cb

    def get_base(
        self,
        role: LevelRole | GroupRole,
        fail_cb: Optional[SyncOrAsyncCallable[[], None]] = None,
    ) -> MsgChecker:
        """根据内部依据和给定等级，生成一个 :class:`MsgChecker` 对象

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param fail_cb: 检查不通过的回调
        :return: 消息事件分级权限检查器
        """
        return MsgChecker(
            role,
            self.owner,
            self.super_users,
            self.white_users,
            self.black_users,
            self.fail_cb if fail_cb is None else fail_cb,
        )

    def get_group(
        self,
        role: LevelRole | GroupRole,
        fail_cb: Optional[SyncOrAsyncCallable[[], None]] = None,
    ) -> GroupMsgChecker:
        """根据内部依据和给定等级，生成一个 :class:`GroupMsgChecker` 对象

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param fail_cb: 检查不通过的回调
        :return: 群聊消息事件分级权限检查器
        """
        return GroupMsgChecker(
            role,
            self.owner,
            self.super_users,
            self.white_users,
            self.black_users,
            self.white_groups,
            self.fail_cb if fail_cb is None else fail_cb,
        )

    def get_private(
        self,
        role: LevelRole,
        fail_cb: Optional[SyncOrAsyncCallable[[], None]] = None,
    ) -> PrivateMsgChecker:
        """根据内部依据和给定等级，生成一个 :class:`PrivateMsgChecker` 对象

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param fail_cb: 检查不通过的回调
        :return: 私聊消息事件分级权限检查器
        """
        return PrivateMsgChecker(
            role,
            self.owner,
            self.super_users,
            self.white_users,
            self.black_users,
            self.fail_cb if fail_cb is None else fail_cb,
        )


class AtMsgChecker(Checker):
    """艾特消息事件检查器"""

    def __init__(
        self,
        qid: int | Literal["all"] | None = None,
        fail_cb: Optional[SyncOrAsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个艾特消息事件检查器

        :param qid: 被艾特的 qq 号。为空则接受所有艾特消息事件；不为空则只接受指定 qid 被艾特的艾特消息事件
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(fail_cb)
        self.qid = qid
        self._hash_tag = qid

    async def check(self, event: Event) -> bool:
        # 不要使用 isinstace，避免通过反射模式注入的 event 依赖产生误判结果
        status = is_msg = False
        try:
            ret, is_msg = event.flag_get(
                self.__class__, self._hash_tag, raise_exc=False, default=(None, False)
            )
            if ret is not None:
                return cast(bool, ret)

            if not event.is_message():
                status = is_msg = False
            else:
                is_msg = True
                event = cast(MessageEvent, event)
                qids = [seg.data["qq"] for seg in event.message if isinstance(seg, AtSegment)]
                if self.qid is None:
                    status = len(qids) > 0
                else:
                    status = any(id == self.qid for id in qids)

            event.flag_set(self.__class__, self._hash_tag, (status, is_msg))
            return status

        finally:
            if not status and is_msg and self.fail_cb is not None:
                await self.fail_cb()



import concurrent.futures
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import threading
from argparse import Namespace
from pathlib import Path

from typing_extensions import Any, Callable

from melobot import __version__
from melobot._run import CLI_LAST_EXIT_CODE, CLI_RUN_ALIVE_FLAG, CLI_RUN_FLAG
from melobot.typ import ExitCode


def main(args: Namespace) -> None:
    Observer, Handler = get_requires()
    observer = Observer()
    reload_signal = threading.Event()
    for path in args.watch:
        observer.schedule(Handler(path, reload_signal), path, recursive=True)

    if not args.entry_file.endswith(".py"):
        entry_str = f"{args.entry_file}.py"
    else:
        entry_str = args.entry_file

    entry_path = Path(entry_str)
    try:
        entry = entry_path.resolve(strict=True).as_posix()
    except FileNotFoundError:
        raise FileNotFoundError(f"不存在的入口文件：{args.entry_file}")

    cmd = [sys.executable, entry]
    cwd = Path.cwd().resolve().as_posix()
    os.environ[CLI_RUN_FLAG] = "1"
    tmp_dir = Path(tempfile.mkdtemp()).resolve()
    os.environ[CLI_RUN_ALIVE_FLAG] = str(
        tmp_dir.joinpath(f"melobot_cli_dev_run_{threading.get_native_id()}.signal")
    )

    try:
        observer.start()
        while True:
            retcode: int
            create_alive_sig()
            pre_handlers: list | None = None
            retcodes: list[int] = []

            with subprocess.Popen(
                cmd,
                env=os.environ,
                cwd=cwd,
                stdin=sys.stdin,
                stdout=sys.stdout,
                stderr=sys.stderr,
            ) as proc:

                if reload_signal.is_set():
                    reload_signal.clear()

                pre_handlers = set_signal_handler(proc, retcodes)
                with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
                    futures: list[concurrent.futures.Future] = [
                        executor.submit(proc.wait),
                        executor.submit(reload_signal.wait),
                    ]
                    _, _ = concurrent.futures.wait(
                        futures, return_when=concurrent.futures.FIRST_COMPLETED
                    )

                    if reload_signal.is_set():
                        clear_alive_sig()
                        proc.wait()
                        retcode = ExitCode.RESTART.value
                    else:
                        # 避免一直等待导致线程池无法关闭
                        reload_signal.set()
                        retcode = proc.returncode

            if len(retcodes):
                retcode = retcodes[0]

            if pre_handlers is not None:
                clear_signal_handler(pre_handlers)

            if CLI_LAST_EXIT_CODE in os.environ:
                os.environ.pop(CLI_LAST_EXIT_CODE)

            if retcode == ExitCode.NORMAL.value:
                break

            if retcode == ExitCode.RESTART.value:
                print("\n>>> [mb-cli] 正在重启 Bot 主程序\n")
                os.environ[CLI_LAST_EXIT_CODE] = str(ExitCode.RESTART.value)
                continue

            if retcode == ExitCode.ERROR.value:
                break

            print()
            print(f">>> [mb-cli] Bot 主程序返回了意料之外的退出码: {retcode}")
            print(">>> [mb-cli] 若提示“已安全停止运行”，则无需关注此警告")
            break

    finally:
        clear_alive_sig()
        observer.stop()
        observer.join()
        shutil.rmtree(tmp_dir)


def create_alive_sig() -> None:
    with open(os.environ[CLI_RUN_ALIVE_FLAG], "wb"):
        pass


def clear_alive_sig() -> None:
    if os.path.exists(os.environ[CLI_RUN_ALIVE_FLAG]):
        os.remove(os.environ[CLI_RUN_ALIVE_FLAG])


def set_signal_handler(proc: subprocess.Popen, rets: list[int]) -> list:
    def signal_handler(*_: Any, **__: Any) -> None:
        # 重入安全
        clear_alive_sig()
        retcode = proc.wait()
        rets.clear()
        rets.append(retcode)

    pre_handlers = []
    pre_handlers.append(signal.getsignal(signal.SIGINT))
    pre_handlers.append(signal.getsignal(signal.SIGTERM))
    if sys.platform == "win32":
        pre_handlers.append(signal.getsignal(signal.SIGBREAK))

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    if sys.platform == "win32":
        signal.signal(signal.SIGBREAK, signal_handler)

    return pre_handlers


def clear_signal_handler(pre_handlers: list) -> None:
    signal.signal(signal.SIGINT, pre_handlers[0])
    signal.signal(signal.SIGTERM, pre_handlers[1])
    if sys.platform == "win32":
        signal.signal(signal.SIGBREAK, pre_handlers[2])


def get_requires() -> tuple[Callable, Callable]:
    try:
        from watchdog.events import FileSystemEvent, FileSystemEventHandler
        from watchdog.observers import Observer

        class Handler(FileSystemEventHandler):
            def __init__(self, path: str, reload_signal: threading.Event) -> None:
                super().__init__()
                self.path = Path(path).resolve()
                self.re_signal = reload_signal

            def _on_event(self, event: FileSystemEvent) -> None:
                e_path = Path(str(event.src_path)).resolve()

                if "__pycache__" in e_path.parts or os.environ[CLI_RUN_ALIVE_FLAG] in e_path.parts:
                    return

                self.re_signal.set()

            def on_moved(self, event: FileSystemEvent) -> None:
                self._on_event(event)

            def on_created(self, event: FileSystemEvent) -> None:
                self._on_event(event)

            def on_modified(self, event: FileSystemEvent) -> None:
                self._on_event(event)

            def on_deleted(self, event: FileSystemEvent) -> None:
                self._on_event(event)

        return Observer, Handler

    except ModuleNotFoundError:
        print(
            f"部分功能需要额外的依赖。安装这些额外依赖：pip install 'melobot[cli]>={__version__}'"
        )
        sys.exit(1)
